/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import { RBTNColor } from '../../types';
export declare class RBTreeNode {
    key: number;
    parent: RBTreeNode;
    left: RBTreeNode;
    right: RBTreeNode;
    color: number;
    constructor(key: number, color?: RBTNColor);
}
export declare const NIL: RBTreeNode;
/**
 * 1. Each node is either red or black.
 * 2. The root node is always black.
 * 3. Leaf nodes are typically NIL nodes and are considered black.
 * 4. Red nodes must have black children.
 * 5. Black balance: Every path from any node to each of its leaf nodes contains the same number of black nodes.
 */
export declare class RedBlackTree {
    constructor();
    protected _root: RBTreeNode;
    get root(): RBTreeNode;
    protected _size: number;
    get size(): number;
    /**
     * The `insert` function inserts a new node with a given key into a red-black tree and fixes any
     * violations of the red-black tree properties.
     * @param {number} key - The key parameter is a number that represents the value to be inserted into
     * the RBTree.
     * @returns The function does not explicitly return anything.
     */
    add(key: number): void;
    /**
     * The `delete` function in TypeScript is used to remove a node with a specific key from a red-black
     * tree.
     * @param {number} key - The `node` parameter is of type `RBTreeNode` and represents the current
     * node being processed in the delete operation.
     * @returns The `delete` function does not return anything. It has a return type of `void`.
     */
    delete(key: number): void;
    isRealNode(node: RBTreeNode | null | undefined): node is RBTreeNode;
    /**
     * The function `getNode` is a recursive depth-first search algorithm that searches for a node with a
     * given key in a red-black tree.
     * @param {number} key - The key parameter is a number that represents the value we are searching for
     * in the RBTree.
     * @param beginRoot - The `beginRoot` parameter is an optional parameter that represents the starting
     * point for the search in the binary search tree. If no value is provided for `beginRoot`, it
     * defaults to the root of the binary search tree (`this.root`).
     * @returns a RBTreeNode.
     */
    getNode(key: number, beginRoot?: RBTreeNode): RBTreeNode;
    /**
     * The function returns the leftmost node in a red-black tree.
     * @param {RBTreeNode} node - The parameter "node" is of type RBTreeNode, which represents a node in
     * a Red-Black Tree.
     * @returns The leftmost node in the given RBTreeNode.
     */
    getLeftMost(node?: RBTreeNode): RBTreeNode;
    /**
     * The function returns the rightmost node in a red-black tree.
     * @param {RBTreeNode} node - The parameter "node" is of type RBTreeNode.
     * @returns the rightmost node in a red-black tree.
     */
    getRightMost(node: RBTreeNode): RBTreeNode;
    /**
     * The function returns the successor of a given node in a red-black tree.
     * @param {RBTreeNode} x - RBTreeNode - The node for which we want to find the successor.
     * @returns the successor of the given RBTreeNode.
     */
    getSuccessor(x: RBTreeNode): RBTreeNode;
    /**
     * The function returns the predecessor of a given node in a red-black tree.
     * @param {RBTreeNode} x - The parameter `x` is of type `RBTreeNode`, which represents a node in a
     * Red-Black Tree.
     * @returns the predecessor of the given RBTreeNode 'x'.
     */
    getPredecessor(x: RBTreeNode): RBTreeNode;
    clear(): void;
    print(beginRoot?: RBTreeNode): void;
    /**
     * The function performs a left rotation on a red-black tree node.
     * @param {RBTreeNode} x - The parameter `x` is a RBTreeNode object.
     */
    protected _leftRotate(x: RBTreeNode): void;
    /**
     * The function performs a right rotation on a red-black tree node.
     * @param {RBTreeNode} x - x is a RBTreeNode, which represents the node that needs to be right
     * rotated.
     */
    protected _rightRotate(x: RBTreeNode): void;
    /**
     * The _fixDelete function is used to rebalance the Red-Black Tree after a node deletion.
     * @param {RBTreeNode} x - The parameter `x` is of type `RBTreeNode`, which represents a node in a
     * red-black tree.
     */
    protected _fixDelete(x: RBTreeNode): void;
    /**
     * The function `_rbTransplant` replaces one node in a red-black tree with another node.
     * @param {RBTreeNode} u - The parameter "u" represents a RBTreeNode object.
     * @param {RBTreeNode} v - The parameter "v" is a RBTreeNode object.
     */
    protected _rbTransplant(u: RBTreeNode, v: RBTreeNode): void;
    /**
     * The `_fixInsert` function is used to fix the red-black tree after an insertion operation.
     * @param {RBTreeNode} k - The parameter `k` is a RBTreeNode object, which represents a node in a
     * red-black tree.
     */
    protected _fixInsert(k: RBTreeNode): void;
}
