/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import type { BTNKey, TreeMultisetNodeNested, TreeMultisetOptions } from '../../types';
import { BinaryTreeDeletedResult, BTNCallback, IterationType } from '../../types';
import { IBinaryTree } from '../../interfaces';
import { AVLTree, AVLTreeNode } from './avl-tree';
export declare class TreeMultisetNode<V = any, N extends TreeMultisetNode<V, N> = TreeMultisetNodeNested<V>> extends AVLTreeNode<V, N> {
    count: number;
    /**
     * The constructor function initializes a BinaryTreeNode object with a key, value, and count.
     * @param {BTNKey} key - The `key` parameter is of type `BTNKey` and represents the unique identifier
     * of the binary tree node.
     * @param {V} [value] - The `value` parameter is an optional parameter of type `V`. It represents the value of the binary
     * tree node. If no value is provided, it will be `undefined`.
     * @param {number} [count=1] - The `count` parameter is a number that represents the number of times a particular value
     * occurs in a binary tree node. It has a default value of 1, which means that if no value is provided for the `count`
     * parameter when creating a new instance of the `BinaryTreeNode` class.
     */
    constructor(key: BTNKey, value?: V, count?: number);
}
/**
 * The only distinction between a TreeMultiset and a AVLTree lies in the ability of the former to store duplicate nodes through the utilization of counters.
 */
export declare class TreeMultiset<V = any, N extends TreeMultisetNode<V, N> = TreeMultisetNode<V, TreeMultisetNodeNested<V>>> extends AVLTree<V, N> implements IBinaryTree<V, N> {
    /**
     * The constructor function for a TreeMultiset class in TypeScript, which extends another class and sets an option to
     * merge duplicated values.
     * @param {TreeMultisetOptions} [options] - An optional object that contains additional configuration options for the
     * TreeMultiset.
     */
    constructor(options?: TreeMultisetOptions);
    private _count;
    get count(): number;
    /**
     * The function creates a new BSTNode with the given key, value, and count.
     * @param {BTNKey} key - The key parameter is the unique identifier for the binary tree node. It is used to
     * distinguish one node from another in the tree.
     * @param {N} value - The `value` parameter represents the value that will be stored in the binary search tree node.
     * @param {number} [count] - The "count" parameter is an optional parameter of type number. It represents the number of
     * occurrences of the value in the binary search tree node. If not provided, the count will default to 1.
     * @returns A new instance of the BSTNode class with the specified key, value, and count (if provided).
     */
    createNode(key: BTNKey, value?: V, count?: number): N;
    /**
     * The `add` function adds a new node to a binary search tree, updating the count if the key already
     * exists, and balancing the tree if necessary.
     * @param {BTNKey | N | null} keyOrNode - The `keyOrNode` parameter can be either a
     * `BTNKey` (which represents the key of the node to be added), a `N` (which represents a
     * node to be added), or `null` (which represents a null node).
     * @param [value] - The `value` parameter represents the value associated with the key that is being
     * added to the binary tree.
     * @param [count=1] - The `count` parameter represents the number of occurrences of the key/value
     * pair that will be added to the binary tree. It has a default value of 1, which means that if no
     * count is specified, the default count will be 1.
     * @returns The function `add` returns a value of type `N | null | undefined`.
     */
    add(keyOrNode: BTNKey | N | null, value?: V, count?: number): N | null | undefined;
    /**
     * The function adds a new node to a binary tree if there is an available slot in the parent node.
     * @param {N | null} newNode - The `newNode` parameter represents the node that needs to be added to
     * the tree. It can be either a node object (`N`) or `null`.
     * @param {N} parent - The `parent` parameter represents the parent node to which the new node will
     * be added as a child.
     * @returns The method `_addTo` returns either the `parent.left`, `parent.right`, or `undefined`.
     */
    _addTo(newNode: N | null, parent: N): N | null | undefined;
    /**
     * The `addMany` function adds multiple keys or nodes to a TreeMultiset and returns an array of the
     * inserted nodes.
     * @param {(BTNKey | null)[] | (N | null)[]} keysOrNodes - An array of keys or nodes to be
     * added to the multiset. Each element can be either a BTNKey or a TreeMultisetNode.
     * @param {V[]} [data] - The `data` parameter is an optional array of values that correspond
     * to the keys or nodes being added to the multiset. It is used to associate additional data with
     * each key or node.
     * @returns The function `addMany` returns an array of `N`, `null`, or `undefined` values.
     */
    addMany(keysOrNodes: (BTNKey | null)[] | (N | null)[], data?: V[]): (N | null | undefined)[];
    /**
     * The `perfectlyBalance` function in TypeScript takes a sorted array of nodes and builds a balanced
     * binary search tree using either a recursive or iterative approach.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when building a balanced binary search tree. It can have two possible
     * values:
     * @returns a boolean value.
     */
    perfectlyBalance(iterationType?: IterationType): boolean;
    /**
     * The `delete` function in a binary search tree deletes a node from the tree and returns the deleted
     * node along with the parent node that needs to be balanced.
     * @param {ReturnType<C>} identifier - The `identifier` parameter is either a
     * `BTNKey` or a generic type `N`. It represents the property of the node that we are
     * searching for. It can be a specific key value or any other property of the node.
     * @param callback - The `callback` parameter is a function that takes a node as input and returns a
     * value. This value is compared with the `identifier` parameter to determine if the node should be
     * included in the result. The `callback` parameter has a default value of
     * `this.defaultOneParamCallback`
     * @param [ignoreCount=false] - A boolean flag indicating whether to ignore the count of the node
     * being deleted. If set to true, the count of the node will not be considered and the node will be
     * deleted regardless of its count. If set to false (default), the count of the node will be
     * decremented by 1 and
     * @returns The method `delete` returns an array of `BinaryTreeDeletedResult<N>` objects.
     */
    delete<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback?: C, ignoreCount?: boolean): BinaryTreeDeletedResult<N>[];
    /**
     * The clear() function clears the contents of a data structure and sets the count to zero.
     */
    clear(): void;
    /**
     * The function swaps the values of two nodes in a binary tree.
     * @param {N} srcNode - The source node that needs to be swapped with the destination node.
     * @param {N} destNode - The `destNode` parameter represents the destination node where the values
     * from `srcNode` will be swapped into.
     * @returns The method is returning the `destNode` after swapping its properties with the `srcNode`.
     */
    protected _swap(srcNode: N, destNode: N): N;
    /**
     * The function sets the value of the "_count" property.
     * @param {number} v - number
     */
    protected _setCount(v: number): void;
}
